#
# This file is part of tracCGit.
#
# Copyright (c) 2012 Ferry Huberts
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from genshi.builder import tag
from genshi.core import Markup
from trac.config import BoolOption, Option
from trac.core import *
from trac.perm import IPermissionRequestor
from trac.util.text import to_unicode
from trac.util.translation import _
from trac.web.api import IRequestFilter, IRequestHandler
from trac.web.chrome import INavigationContributor, ITemplateProvider, \
    add_stylesheet
import re
import urllib
import urllib2



class CGitModule(Component):
    implements(IPermissionRequestor, INavigationContributor, IRequestFilter, IRequestHandler, ITemplateProvider)

    cgit_redirect = BoolOption('cgit', 'cgit_redirect', default=False, doc='Redirect to cgit project page when true, embed in trac interface when false')
    cgit_url = Option('cgit', 'cgit_url', doc='URL to cgit project page. Example: /cgit/projectName')

    patterns = [
        (r'^.*<body(|\s+[^>]*)>', ''),
        (r'<\/body(|\s+[^>])*>.*', ''),
        (r'(<td(|\s+[^>]+?)\s+class=["\'](logo|main)["\'][^>]*>.*?<\/td>)', '<!-- \\1 --><td></td>'),
        (r'(<tr(|\s+[^>]*)>\s*<td(|\s+[^>]+?)\s+class=["\']sub["\'][^>]*>.*?<\/td>\s*<\/tr>)', '<!-- \\1 -->')
    ]

    patterns = [(re.compile(pat, re.S | re.I | re.U), rep) for pat, rep in patterns]


    #
    # Utilities
    #

    def fixCGitUrl(self):
        if not self.cgit_url:
            raise TracError('You must configure a cgit URL in trac.ini')

        return self.cgit_url.strip().rstrip('/')


    def fixCGitUrlBase(self):
        return self.fixCGitUrl().rpartition('/')[0].rstrip('/')


    #
    # IPermissionRequestor
    #
    # Extension point interface for components that define actions. 
    #

    def get_permission_actions(self):
        yield 'BROWSER_VIEW'


    #
    # INavigationContributor
    #
    # Extension point interface for components that contribute items to the
    # navigation. 
    #

    def get_active_navigation_item(self, req):
        return 'browser'


    def get_navigation_items(self, req):
        if ('BROWSER_VIEW' in req.perm) and self.cgit_url:
            yield 'metanav', 'cgit', tag.a('CGit', href=self.fixCGitUrl())


    #
    # IRequestFilter
    #
    # Extension point interface for components that want to filter HTTP
    # requests, before and/or after they are processed by the main handler. 
    #

    def pre_process_request(self, req, handler):
        if self.cgit_url:
            self.env.log.debug('pre_process_request: req             =%r', req)
            self.env.log.debug('pre_process_request: req.path_info   =%s', req.path_info)
            self.env.log.debug('pre_process_request: req.query_string=%s', req.query_string)

            query_string = req.query_string

            if req.path_info.startswith('/browser'):
                cgitUrl = re.sub('^/browser', self.fixCGitUrl(), req.path_info)
            elif req.path_info.startswith('/changeset'):
                if query_string.find('id=') != -1:
                    raise TracError('Internal error: request should have already been redirected, req=%r' % req)

                sha1 = re.sub('^/changeset/*', '', req.path_info)
                sha1 = sha1.partition('/')[0];
                if sha1 != '':
                    sha1 = 'id=' + sha1
                    if query_string == '':
                        query_string = sha1
                    else:
                        query_string = sha1 + '&' + query_string

                cgitUrl = req.base_path + '/browser/commit/'
                if query_string != '':
                    cgitUrl = cgitUrl + '?' + query_string

                self.env.log.debug('pre_process_request: redirect to %s' % cgitUrl)
                req.redirect(cgitUrl)
                return handler

            else:
                    return handler

            req.perm.assert_permission('BROWSER_VIEW')

            self.env.log.debug('pre_process_request: cgitUrl         =%s', cgitUrl)
            if query_string != '':
                cgitUrl = cgitUrl + '?' + query_string
                self.env.log.debug('pre_process_request: cgitUrl         =%s', cgitUrl)

            if re.match('^/browser/(atom|patch|plain|snapshot)(|/.*)$', req.path_info) != None:
                self.env.log.debug('pre_process_request: redirect plain file to %s' % cgitUrl)
                req.redirect(cgitUrl)
                return handler

            if (not self.cgit_redirect):
                self.env.log.debug('pre_process_request: embed %s' % cgitUrl)
                req.cgitUrl = cgitUrl
                return self

            self.env.log.debug('pre_process_request: redirect to %s' % cgitUrl)
            req.redirect(cgitUrl)
            return handler


        return handler


    def post_process_request(self, req, template, data, content_type):
        return (template, data, content_type)


    #
    # IRequestHandler
    #
    # Extension point interface for request handlers. 
    #

    def match_request(self, req):
        self.env.log.debug('match_request: req                   =%r', req)
        return req.path_info.startswith('/browser') or req.path_info.startswith('/changeset')


    def process_request(self, req):
        self.env.log.debug('process_request: req                 =%r', req)

        req.perm.assert_permission('BROWSER_VIEW')

        if self.cgit_redirect:
            raise TracError('process_request was called with redirect=True')

        if not self.cgit_url:
            raise TracError('You must configure a cgit URL in trac.ini')
        
        if not req.cgitUrl:
            raise TracError('No req.cgitUrl found')

        self.env.log.debug('process_request: req.path_info       =%s', req.path_info)
        self.env.log.debug('process_request: req.cgitUrl         =%s', req.cgitUrl)

        urlf = urllib2.urlopen('http://localhost' + req.cgitUrl)
        page = urlf.read()

        page = to_unicode(page)
        for pat, rep in self.patterns:
            page = pat.sub(rep, page)

        pat = re.compile(r'(href=["\'])' + self.fixCGitUrl(), re.S | re.I | re.U)
        page = pat.sub('\\1' + req.href.browser(), page)

        pat = re.compile(r'(href=["\'])' + self.fixCGitUrlBase() + '((|/)["\'])', re.S | re.I | re.U)
        page = pat.sub('\\1' + req.href.browser() + '\\2', page)

        data = {
            'cgit_page': Markup(page)
        }

        add_stylesheet(req, 'cgit/cgit.css')
        return 'cgit.html', data, urlf.info().type


    #
    # ITemplateProvider methods
    #
    # Extension point interface for components that provide their own
    #   ClearSilver templates and accompanying static resources.
    #

    def get_htdocs_dirs(self):
        from pkg_resources import resource_filename
        return [('cgit', resource_filename(__name__, 'htdocs'))]


    def get_templates_dirs(self):
        from pkg_resources import resource_filename
        return [resource_filename(__name__, 'templates')]
